----------------------------------------------------------------------------------
-- COPYRIGHT (c) 2020 ALL RIGHT RESERVED
--
-- COMPANY:					Ruhr-University Bochum, Chair for Security Engineering
-- AUTHOR:					Jan Richter-Brockmann
--
-- CREATE DATE:			    04/05/2020
-- LAST CHANGES:            04/05/2020
-- MODULE NAME:			    BIKE_HASH_ERROR
--
-- REVISION:				1.00 - File created.
--
-- LICENCE: 				Please look at licence.txt
-- USAGE INFORMATION:	    Please look at readme.txt. If licence.txt or readme.txt
--							are missing or	if you have questions regarding the code
--							please contact Tim Gneysu (tim.gueneysu@rub.de) and
--                          Jan Richter-Brockmann (jan.richter-brockmann@rub.de)
--
-- THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY 
-- KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
-- PARTICULAR PURPOSE.
----------------------------------------------------------------------------------



-- IMPORTS
----------------------------------------------------------------------------------
LIBRARY IEEE;
    USE IEEE.STD_LOGIC_1164.ALL;
    USE IEEE.NUMERIC_STD.ALL;
    USE IEEE.MATH_REAL.ALL;

LIBRARY UNISIM;
    USE UNISIM.VCOMPONENTS.ALL;
    
LIBRARY work;
    USE work.BIKE_SETTINGS.ALL;



-- ENTITY
----------------------------------------------------------------------------------
ENTITY BIKE_HASH_ERROR IS
	PORT (  
        CLK                 : IN  STD_LOGIC;
        -- CONTROL PORTS ---------------	
        RESET               : IN  STD_LOGIC;
        HASH_EN             : IN  STD_LOGIC;
        HASH_DONE           : IN  STD_LOGIC;
        -- ERROR BRAM ------------------
        SAMPLE_EN           : IN  STD_LOGIC;
        ERROR0_WREN         : IN  STD_LOGIC;
        ERROR1_WREN         : IN  STD_LOGIC;
        ERROR_ADDR          : IN  STD_LOGIC_VECTOR(LOG2(R_BLOCKS)-1 DOWNTO 0);
        ERROR_DIN           : IN  STD_LOGIC_VECTOR(31 DOWNTO 0);
        -- HASH ------------------------
        HASH_M              : OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
        HASH_VALID          : OUT STD_LOGIC;
        HASH_RDY            : IN  STD_LOGIC
    );
END BIKE_HASH_ERROR;



-- ARCHITECTURE
----------------------------------------------------------------------------------
ARCHITECTURE Structural OF BIKE_HASH_ERROR IS



-- CONSTANTS
----------------------------------------------------------------------------------
--CONSTANT OVERHANG        : INTEGER := R_BITS - 32*(R_BLOCKS-1);   -- use this setting if e represented as one polynomial should be hashed
CONSTANT OVERHANG        : INTEGER := 8 * CEIL(R_BITS - 32*(R_BLOCKS-1), 8);    -- use this setting if e0 || e1 should be hashed (represented in bytes)



-- STATES
----------------------------------------------------------------------------------
TYPE STATES IS (S_RESET, S_WRITE_ERROR, S_HASH_FIRST_PART, S_HASH_SWITCH, S_HASH_SECOND_PART, S_DONE);
SIGNAL STATE : STATES := S_RESET;



-- SIGNALS
----------------------------------------------------------------------------------
-- COUNTER
SIGNAL CNT_ADDR_EN, CNT_ADDR_RST                        : STD_LOGIC;
SIGNAL CNT_ADDR_OUT                                     : STD_LOGIC_VECTOR(LOG2(R_BLOCKS)-1 DOWNTO 0);

-- BRAM PORTS HASH
SIGNAL SAMPLE_EN_GATED                                  : STD_LOGIC;
SIGNAL ERROR0_RDEN, ERROR1_RDEN                         : STD_LOGIC;
SIGNAL ERROR0_DOUT, ERROR1_DOUT                         : STD_LOGIC_VECTOR(31 DOWNTO 0);
SIGNAL ERROR0_ADDR, ERROR1_ADDR                         : STD_LOGIC_VECTOR(LOG2(R_BLOCKS)-1 DOWNTO 0);
SIGNAL ADDR_E0, ADDR_E1                                 : STD_LOGIC_VECTOR(LOG2(R_BLOCKS)-1 DOWNTO 0);


-- SHA
SIGNAL SHA_ENABLE, SHA_RESET                            : STD_LOGIC;
SIGNAL SHA_M_VALID, SHA_HASH_VALID                      : STD_LOGIC;


-- CONTROLLING
SIGNAL SWITCH_POLY, SWITCH_POLY_D, SECOND_PART          : STD_LOGIC;
SIGNAL FIRST_M, SECOND_M, COMP_M                        : STD_LOGIC_VECTOR(31 DOWNTO 0);
SIGNAL SECOND_M_PRE, SECOND_M_SWITCH, SECOND_M_SW       : STD_LOGIC_VECTOR(31 DOWNTO 0);
SIGNAL INT_ERROR1                                       : STD_LOGIC_VECTOR(OVERHANG-1 DOWNTO 0);

SIGNAL SHA_ENABLE_GETED : STD_LOGIC;


-- STRUCTURAL
----------------------------------------------------------------------------------
BEGIN

    -- MEMORY --------------------------------------------------------------------
    BRAM_E : ENTITY work.BIKE_BRAM
    GENERIC MAP (
        OUTPUT_BRAM     => 1
    )
    PORT MAP (
        -- CONTROL PORTS ----------------
        CLK             => CLK,     
        RESET           => RESET,
        SAMPLING        => SAMPLE_EN_GATED,
        -- SAMPLING --------------------
        REN0_SAMP       => ERROR0_RDEN,
        REN1_SAMP       => ERROR1_RDEN,
        WEN0_SAMP       => ERROR0_WREN,
        WEN1_SAMP       => ERROR1_WREN,
        ADDR0_SAMP      => ERROR0_ADDR,
        ADDR1_SAMP      => ERROR1_ADDR,
        DOUT0_SAMP      => ERROR0_DOUT,
        DOUT1_SAMP      => ERROR1_DOUT,
        DIN0_SAMP       => ERROR_DIN,
        DIN1_SAMP       => ERROR_DIN,
        -- COMPUTATION -----------------
        WEN0            => '0',
        WEN1            => '0',
        REN0            => '0',
        REN1            => '0',
        ADDR0           => (OTHERS => '0'),
        ADDR1           => (OTHERS => '0'),
        DOUT0           => OPEN,
        DOUT1           => OPEN,
        DIN0            => (OTHERS => '0'),
        DIN1            => (OTHERS => '0')
    );
    
    ERROR0_ADDR <= ERROR_ADDR WHEN SAMPLE_EN = '1' ELSE ADDR_E0; 
    ERROR1_ADDR <= ERROR_ADDR WHEN SAMPLE_EN = '1' ELSE ADDR_E1; 
    
    SAMPLE_EN_GATED <= '1' WHEN SAMPLE_EN = '1' ELSE HASH_RDY;
    ERROR0_RDEN <= ERROR0_WREN WHEN SAMPLE_EN = '1' ELSE HASH_RDY;
    ERROR1_RDEN <= ERROR1_WREN WHEN SAMPLE_EN = '1' ELSE HASH_RDY;
    ------------------------------------------------------------------------------
    
    
    
    -- DATA REORDERING -----------------------------------------------------------
    ADDR_E0 <= CNT_ADDR_OUT;
    ADDR_E1 <= (OTHERS => '0') WHEN SWITCH_POLY = '1' ELSE STD_LOGIC_VECTOR(UNSIGNED(CNT_ADDR_OUT)+1);
    
    -- input data need to be rearranged as our memory layout differs that from the reference implementation
    HASH_M  <= FIRST_M WHEN SECOND_PART = '0' ELSE COMP_M;

    -- reordering for e0
    FIRST_M         <= ERROR0_DOUT(7 DOWNTO 0) & ERROR0_DOUT(15 DOWNTO 8) & ERROR0_DOUT(23 DOWNTO 16) & ERROR0_DOUT(31 DOWNTO 24);
    
    -- reordering for switching between e0 and e1
    SECOND_M_SWITCH <= ERROR1_DOUT(32-OVERHANG-1 DOWNTO 0) & ERROR0_DOUT(OVERHANG-1 DOWNTO 0);
    SECOND_M_SW     <= SECOND_M_SWITCH(7 DOWNTO 0) & SECOND_M_SWITCH(15 DOWNTO 8) & SECOND_M_SWITCH(23 DOWNTO 16) & SECOND_M_SWITCH(31 DOWNTO 24);
    
    -- reordering for e1
    SECOND_M_PRE    <= ERROR1_DOUT(32-OVERHANG-1 DOWNTO 0) & INT_ERROR1;    
    SECOND_M        <= SECOND_M_PRE(7 DOWNTO 0) & SECOND_M_PRE(15 DOWNTO 8) & SECOND_M_PRE(23 DOWNTO 16) & SECOND_M_PRE(31 DOWNTO 24);
    
    COMP_M          <= SECOND_M_SW WHEN SWITCH_POLY_D = '1' ELSE SECOND_M;
    
    -- store higher bits in a register
    REG_E2_PART : ENTITY work.RegisterFDRE GENERIC MAP(SIZE => OVERHANG)
    PORT MAP(D => ERROR1_DOUT(31 DOWNTO 32-OVERHANG), Q => INT_ERROR1, CLK => CLK, EN => SECOND_PART, RST => RESET);
    
    REG_SWITCH_POLY : PROCESS(CLK)
    BEGIN
        IF RISING_EDGE(CLK) THEN
            SWITCH_POLY_D <= SWITCH_POLY;
        END IF;
    END PROCESS;
    
    REG_M_VALID : PROCESS(CLK)
    BEGIN
        IF RISING_EDGE(CLK) THEN
            SHA_M_VALID <= HASH_RDY;
        END IF;
    END PROCESS;
    
    HASH_VALID <= SHA_M_VALID;
    ------------------------------------------------------------------------------  
    
 
    -- COUNTER -------------------------------------------------------------------
    CNT_ADDR_EN <= HASH_RDY;
    CNT_ADDR : ENTITY work.BIKE_counter_inc GENERIC MAP(SIZE => LOG2(R_BLOCKS), MAX_VALUE => R_BLOCKS-1)
    PORT MAP(CLK => CLK, EN => CNT_ADDR_EN, RST => CNT_ADDR_RST, CNT_OUT => CNT_ADDR_OUT);
    ------------------------------------------------------------------------------
    
    
    -- FINITE STATE MACHINE ------------------------------------------------------
    FSM : PROCESS(CLK)
    BEGIN
        IF RISING_EDGE(CLK) THEN
            CASE STATE IS
                
                ----------------------------------------------
                WHEN S_RESET        =>
                    -- COUNTER ---------
                    CNT_ADDR_RST    <= '1';
                    
                    -- HASH FUNCTION ---
                    SHA_RESET       <= '1';
                    SHA_ENABLE      <= '0';
                    
                    SWITCH_POLY     <= '0';
                    SECOND_PART     <= '0';
                                 
                    
                    -- TRANSITION ------
                    IF (SAMPLE_EN = '1') THEN
                        STATE       <= S_WRITE_ERROR;
                    ELSE
                        STATE       <= S_RESET;
                    END IF;
                ----------------------------------------------

                ----------------------------------------------
                WHEN S_WRITE_ERROR  =>
                    -- COUNTER ---------
                    CNT_ADDR_RST    <= '1';
                    
                    -- HASH FUNCTION ---
                    SHA_RESET       <= '1';
                    SHA_ENABLE      <= '0';
                    
                    SWITCH_POLY     <= '0';
                    SECOND_PART     <= '0';
                                                                                
                    -- TRANSITION ------
                    IF (HASH_EN = '1') THEN
                        STATE       <= S_HASH_FIRST_PART;
                    ELSE
                        STATE       <= S_WRITE_ERROR;
                    END IF;
                ----------------------------------------------

                ----------------------------------------------
                WHEN S_HASH_FIRST_PART  =>
                    -- COUNTER ---------
                    CNT_ADDR_RST    <= '0';
                    
                    -- HASH FUNCTION ---
                    SHA_RESET       <= '0';
                    SHA_ENABLE      <= '1';
                    
                    SWITCH_POLY     <= '0';
                    SECOND_PART     <= '0';
                                                                                
                    -- TRANSITION ------
                    IF (UNSIGNED(CNT_ADDR_OUT) = R_BLOCKS-3) THEN 
                        STATE       <= S_HASH_SWITCH;
                    ELSE
                        STATE       <= S_HASH_FIRST_PART;                      
                    END IF;
                ----------------------------------------------

                ----------------------------------------------
                WHEN S_HASH_SWITCH         =>
                    -- COUNTER ---------
                    CNT_ADDR_RST    <= '0';
                    
                    -- HASH FUNCTION ---
                    SHA_RESET       <= '0';
                    SHA_ENABLE      <= '1';
                    
                    SWITCH_POLY     <= '1';
                    SECOND_PART     <= '0';
                                  
                    -- TRANSITION ------
                    IF HASH_RDY = '1' THEN
                        STATE       <= S_HASH_SECOND_PART;
                    ELSE
                        STATE       <= S_HASH_SWITCH;
                    END IF;
                ----------------------------------------------

                ----------------------------------------------
                WHEN S_HASH_SECOND_PART         =>
                    -- COUNTER ---------
                    CNT_ADDR_RST    <= '0'; 

                    -- HASH FUNCTION ---
                    SHA_RESET       <= '0';
                    SHA_ENABLE      <= '1';
                    
                    SWITCH_POLY     <= '0';
                    SECOND_PART     <= '1';
                                                                                
                    -- TRANSITION ------
                    IF (HASH_DONE = '1') THEN
                        STATE           <= S_DONE;
                    ELSE
                        STATE           <= S_HASH_SECOND_PART;
                    END IF;
                ----------------------------------------------    
                
                ----------------------------------------------
                WHEN S_DONE         =>
                    -- COUNTER ---------
                    CNT_ADDR_RST    <= '1'; 
                    
                    -- HASH FUNCTION ---
                    SHA_RESET       <= '1';
                    SHA_ENABLE      <= '0';
                    
                    SWITCH_POLY     <= '0';
                    SECOND_PART     <= '0';
                                                            
                    -- TRANSITION ------
                    IF (RESET = '1') THEN
                        STATE       <= S_RESET;
                    ELSE
                        STATE       <= S_DONE;
                    END IF;
                ----------------------------------------------                                              
                                
            END CASE;
        END IF;
    END PROCESS;    
    ------------------------------------------------------------------------------

END Structural;
